/* The PURE parser.  -*- C++ -*- */

// Tell bison that we want a C++ parser.
/* NOTE: We require at least bison 2.1a here, since the C++ parser skeleton
   changed several times, and the newer versions are not compatible with bison
   2.1 and earlier. :( */
%skeleton "lalr1.cc"
%require "2.1a"
%defines

/* The operator section syntax causes a number of shift/reduce conflicts, we
   take care of these here. */
%expect 5

%{
#include <iostream>
#include <string>
#include "parserdefs.hh"
#include "printer.hh"
#include "util.hh"

#define restricted_action(task,cleanup) \
  try { if (interp.nerrs == 0) {task;} else {cleanup;} } \
  catch (err &e) { error(yyloc, e.what()); } \
  interp.nerrs = 0;

#define action(task,cleanup) \
  try { if (interp.restricted) throw err("operation not implemented"); \
        else if (interp.nerrs == 0) {task;} else {cleanup;} } \
  catch (err &e) { error(yyloc, e.what()); } \
  interp.nerrs = 0;

#define parser_action(task,cleanup) \
  try { task; }	catch (err &e) { cleanup; error(yyloc, e.what()); YYERROR; }

using namespace std;

class interpreter;
%}

// The parsing context.
%parse-param { interpreter& interp }
%lex-param   { interpreter& interp }

// Location tracking.
%locations
%initial-action
{
  static string stdin_source = "<stdin>";
  // Initialize the initial location.
  if (interp.source.empty())
    @$.begin.filename = @$.end.filename = &stdin_source;
  else
    @$.begin.filename = @$.end.filename = &interp.source;
};

// Enable parser tracing and verbose error messages.
%debug
%error-verbose

// Symbols.

%union
{
  char    cval;
  int32_t ival;
  double  dval;
  mpz_t  *zval;
  string *sval;
  char   *csval;
  expr   *xval;
  exprl  *xlval;
  exprll *xllval;
  OpStack *opstk;
  rule   *rval;
  rulel  *rlval;
  rhs_info *rhsval;
  rule_info *rinfo;
  pat_rule_info *prinfo;
  list<string> *slval;
  comp_clause_list *clauselval;
  comp_clause *clauseval;
  fix_t   fix;
  sym_info *info;
};

%{
#include "lexerdefs.hh"
#include "interpreter.hh"
static int extern_priv;
static void mangle_fname(string& name);
%}

%token		PRIVATE	"private"
%token		PUBLIC	"public"
%token		NONFIX	"nonfix"
%token		OUTFIX	"outfix"
%token <fix>	FIX	"fixity"

%token		DEF	"def"
%token		CONST	"const"
%token		LET	"let"
%token		CASE	"case"
%token		OF	"of"
%token		END	"end"
%token		IF	"if"
%token		THEN	"then"
%token		ELSE	"else"
%token		OTHERWISE "otherwise"
%token		WHEN	"when"
%token		WITH	"with"
%token		USING	"using"
%token		NAMESPACE "namespace"
%token		EXTERN	"extern"

%token <xval>	LO	"left outfix operator"
%token <xval>	RO	"right outfix operator"
%token <xval>	NA	"infix operator"
%token <xval>	LT	"infixl operator"
%token <xval>	RT	"infixr operator"
%token <xval>	PR	"prefix operator"
%token <xval>	PO	"postfix operator"

%right		MAPSTO
%left		WHEN WITH ELSE

%token		EOFTOK 0 "end of file"
%token		ERRTOK  "invalid character"
%token		BADTOK  "bad token"
%token		MAPSTO  "->"
%token <sval>	ID	"identifier"
%token <csval>	STR	"string"
%token <ival>	INT	"integer"
%token <zval>	BIGINT	"bigint"
%token <zval>	CBIGINT	"converted bigint"
%token <dval>	DBL	"floating point number"
%token <ival>	TAG	"type tag"
%type  <sval>	name fname optalias ctype
%type  <slval>	ids names fnames ctypes opt_ctypes
%type  <info>	fixity
%type  <xval>	expr prim
%type  <opstk>  simple  "simple expression"
%type  <rhsval> rhs qual_rhs
%type  <xlval>	args lhs row
%type  <xllval>	rows row_list
%type  <clauselval>  comp_clauses comp_clause_list
%type  <clauseval>  comp_clause
%type  <rinfo>	rules rulel
%type  <prinfo>	pat_rules pat_rulel
%type  <rval>	simple_rule
%type  <rlval>	rule simple_rules simple_rulel

%destructor { delete $$; } ID fixity expr simple prim
  comp_clauses comp_clause_list rows row_list row args lhs rhs qual_rhs
  rules rulel rule pat_rules pat_rulel simple_rules simple_rulel simple_rule
  ids fnames fname names name optalias opt_ctypes ctypes ctype
%destructor { mpz_clear(*$$); free($$); } BIGINT CBIGINT
%destructor { free($$); } STR
%printer { debug_stream() << *$$; } ID name fname optalias ctype expr
  prim args lhs rule simple_rules simple_rulel simple_rule
%printer { debug_stream() << $$->r; } rhs qual_rhs
%printer { debug_stream() << $$->e; } rules rulel
%printer { debug_stream() << $$->rl; } pat_rules pat_rulel
%printer { debug_stream() << $$; }  INT DBL STR
%printer { char *s = mpz_get_str(NULL, 10, *$$);
           debug_stream() << s; free(s); }  BIGINT CBIGINT

%%

// Pure grammar.

%start source;

/* An item in the source can either be an expression, a variable binding
   (let), a rewriting rule or an extern, fixity or using declaration. These
   must all be terminated with a semicolon. Current error recovery is
   panic-mode, which doesn't work too well. :( */

source
: /* empty */
| source ';'
| source item ';'
| error ';'		{ interp.nerrs = yyerrstatus_ = 0; }
;

item
: expr
{ restricted_action(interp.exec($1), delete $1); }
| LET simple_rule
{ action(interp.define($2), delete $2); }
| CONST simple_rule
{ action(interp.define_const($2), delete $2); }
| DEF simple_rule
{ action(interp.add_macro_rule($2), delete $2); }
| rule
{ rulel *rl = 0;
  action(interp.add_rules(interp.globenv,
  (rl = interp.default_lhs(interp.last, $1)), true), if (rl) delete rl); }
| fixity
/* Lexical tie-in: We need to tell the lexer that we're defining new operator
   symbols (interp.declare_op = true) instead of searching for existing ones
   in the symbol table. */
{ if ($1->special && $1->fix != nonfix && $1->fix != outfix &&
      $1->prec >= PREC_MAX) {
    error(yylloc, "invalid fixity declaration"); $1->prec = 0;
  }
  if ($1->fix == nonfix || $1->fix == outfix || $1->prec < PREC_MAX)
    interp.declare_op = true; }
  ids
{ interp.declare_op = false;
  action(interp.declare($1->priv, $1->prec, $1->fix, $3), delete $3);
  delete $1; }
| USING fnames
{ action(interp.run(*$2), {}); delete $2; }
| NAMESPACE name
{ size_t k = $2->rfind("::");
  if (k != string::npos &&
      interp.namespaces.find($2->substr(0, k)) == interp.namespaces.end()) {
    error(yylloc, "unknown namespace '"+$2->substr(0, k)+"'");
    interp.symtab.current_namespace->clear();
  } else {
    interp.namespaces.insert(*$2);
    delete interp.symtab.current_namespace;
    interp.symtab.current_namespace = $2;
  }
}
| NAMESPACE
{ interp.symtab.current_namespace->clear(); }
| USING NAMESPACE names
{ interp.using_namespaces($3); }
| USING NAMESPACE
{ interp.using_namespaces(); }
| PUBLIC EXTERN { extern_priv = 0; } prototypes
| PRIVATE EXTERN { extern_priv = 1; } prototypes
| EXTERN { extern_priv = -1; } prototypes
;

fixity
: FIX INT		{ $$ = new sym_info(true, false,$2,$1); }
| OUTFIX		{ $$ = new sym_info(true, false,PREC_MAX,outfix); }
| NONFIX		{ $$ = new sym_info(true, false,PREC_MAX,nonfix); }
| PUBLIC FIX INT	{ $$ = new sym_info(true, false,$3,$2); }
| PUBLIC OUTFIX		{ $$ = new sym_info(true, false,PREC_MAX,outfix); }
| PUBLIC NONFIX		{ $$ = new sym_info(true, false,PREC_MAX,nonfix); }
| PUBLIC		{ $$ = new sym_info(false, false,PREC_MAX,infix); }
| PRIVATE FIX INT	{ $$ = new sym_info(true, true,$3,$2); }
| PRIVATE OUTFIX	{ $$ = new sym_info(true, true,PREC_MAX,outfix); }
| PRIVATE NONFIX	{ $$ = new sym_info(true, true,PREC_MAX,nonfix); }
| PRIVATE		{ $$ = new sym_info(false, true,PREC_MAX,infix); }
;

ids
: ID
{ $$ = new list<string>; $$->push_back(*$1); delete $1; }
| ids ID
{ $$ = $1; $$->push_back(*$2); delete $2; }
;

fnames
: fname
{ $$ = new list<string>; $$->push_back(*$1); delete $1; }
| fnames ',' fname
{ $$ = $1; $$->push_back(*$3); delete $3; }
;

fname
: ID			{ $$ = $1; mangle_fname(*$$); }
| STR			{ char *s = fromutf8($1); free($1);
			  $$ = new string(s); free(s); }
;

names
: name
{ $$ = new list<string>; $$->push_back(*$1); delete $1; }
| names ',' name
{ $$ = $1; $$->push_back(*$3); delete $3; }
;

name
: ID
{ $$ = $1; if ($$->compare(0, 2, "::") == 0) $$->erase(0, 2); }
| STR
{ $$ = new string($1); free($1);
  if ($$->compare(0, 2, "::") == 0) $$->erase(0, 2); }
;

prototypes
: prototype
| prototypes ',' prototype
;

prototype
: ctype ID '(' opt_ctypes ')' optalias
{ action(interp.declare_extern(extern_priv, *$2, *$1, *$4, false, 0, *$6), {});
  delete $1; delete $2; delete $4; delete $6; }
;

opt_ctypes
: /* empty */
{ $$ = new list<string>; }
| ctypes
;

ctypes
: ctype optname
{ $$ = new list<string>; $$->push_back(*$1); delete $1; }
| ctypes ',' ctype optname
{ $$ = $1; $$->push_back(*$3); delete $3; }
;

ctype
: ID
| ctype '*'		{ $$ = $1; *$$ += "*"; }
;

optname
: /* empty */
| ID			{ delete $1; }
;

optalias
: /* empty */		{ $$ = new string; }
| '=' ID		{ $$ = $2; }
;

/* Expression types are either simple expression (see below) or one of the
   special lambda, case, when/with and if-then-else constructs. Lambda binds
   most weakly, followed by when/with and case, followed by if-then-else and
   simple expressions. */

expr
: simple
{ parser_action($$ = interp.mksimple_expr($1), delete $1); }
| '\\' args MAPSTO expr
{ try { $$ = interp.mklambda_expr($2, $4); }
  catch (err &e) { interp.error(yyloc, e.what()); $$ = new expr; } }
| CASE expr OF pat_rules END
{ $$ = interp.mkcase_expr($2, new rulel($4->rl)); delete $4; }
| expr WHEN simple_rules END
{ try { $$ = interp.mkwhen_expr($1, $3); }
  catch (err &e) { interp.error(yyloc, e.what()); $$ = new expr; } }
| expr WITH rules END
{ $$ = interp.mkwith_expr($1, new env($3->e)); delete $3; }
| IF expr THEN expr ELSE expr
{ $$ = interp.mkcond_expr($2, $4, $6); }
;

args
: prim
{ $$ = new exprl; $$->push_back(*$1); delete $1; }
| args prim
{ $$ = $1; $$->push_back(*$2); delete $2; }
;

/* Simple expressions (infix, prefix, postfix, applications). These are parsed
   as a sequence of primaries here, which is then rearranged using a separate
   operator precedence parser. This enables us to deal with an unlimited
   number of precedence levels. */

simple
: prim			{ $$ = (new OpStack())->push_arg($1); }
| PR			{ $$ = (new OpStack())->push_op($1); }
| simple prim		{ $$ = $1->push_arg($2); }
| simple PO		{ $$ = $1->push_op($2); }
| simple PR		{ $$ = $1->push_op($2); }
| simple LT		{ $$ = $1->push_op($2); }
| simple RT		{ $$ = $1->push_op($2); }
| simple NA		{ $$ = $1->push_op($2); }
;

/* Primary expressions. */

prim
: ID			{ try { $$ = interp.mksym_expr($1); }
			  catch (err &e) {
			    interp.error(yyloc, e.what());
			    $$ = interp.mksym_expr(new string("_"));
			  } }
| ID TAG		{ try { $$ = interp.mksym_expr($1, $2); }
			  catch (err &e) {
			    interp.error(yyloc, e.what());
			    $$ = interp.mksym_expr(new string("_"));
			  } }
| ID '@' prim		{ try { $$ = interp.mkas_expr($1, $3); }
			  catch (err &e) {
			    interp.error(yyloc, e.what());
			    $$ = $3;
			  } }
| INT			{ $$ = new expr(EXPR::INT, $1); }
| CBIGINT		{ $$ = new expr(EXPR::BIGINT, *$1, true); free($1); }
| BIGINT		{ $$ = new expr(EXPR::BIGINT, *$1); free($1); }
| DBL			{ $$ = new expr(EXPR::DBL, $1); }
| STR			{ $$ = new expr(EXPR::STR, $1); }
| LO expr RO		{ int32_t g = interp.symtab.sym($1->tag()).g;
			  assert(g != 0);
			  if (g == $3->tag())
			    $$ = interp.mkexpr($1, $2);
			  else {
			    string id = interp.symtab.sym($3->tag()).s;
			    string rid = interp.symtab.sym(g).s;
			    string msg = "syntax error, unexpected '"+id+
			      "', expecting '"+rid+"'";
			    interp.error(yyloc, msg);
			    YYERROR;
			  } }
| '{' rows '}'		{ $$ = new expr(EXPR::MATRIX, $2); }
| '{' expr '|' comp_clauses '}'
			{ $$ = interp.mkmatcomp_expr($2, $4); }
| '[' expr ']'		{ $$ = interp.mklist_expr($2); }
| '[' ']'		{ $$ = new expr(interp.symtab.nil_sym().f); }
/*
| '[' expr ';' comp_clauses ']'
			{ interp.warning(yyloc,
			    "warning: deprecated comprehension syntax");
			  $$ = interp.mklistcomp_expr($2, $4); }
*/
| '[' expr '|' comp_clauses ']'
			{ $$ = interp.mklistcomp_expr($2, $4); }
| '(' expr ')'		{ $$ = $2;
			  if ($$->is_pair()) $$->flags() |= EXPR::PAREN; }
| '(' ')'		{ $$ = new expr(interp.symtab.void_sym().f); }
| '(' NA ')'		{ $$ = $2; }
| '(' LT ')'		{ $$ = $2; }
| '(' RT ')'		{ $$ = $2; }
| '(' PR ')'		{ $$ = $2; }
| '(' PO ')'		{ $$ = $2; }
| '(' LO RO ')'		{ int32_t g = interp.symtab.sym($2->tag()).g;
			  assert(g != 0);
			  if (g == $3->tag())
			    $$ = $2;
			  else {
			    string id = interp.symtab.sym($3->tag()).s;
			    string rid = interp.symtab.sym(g).s;
			    string msg = "syntax error, unexpected '"+id+
			      "', expecting '"+rid+"'";
			    interp.error(yyloc, msg);
			    YYERROR;
			  } }

/* Left sections. These cause a number of shift/reduce conflicts with the
   rules for simple above. */

| '(' simple NA ')'
{ expr *x; parser_action(x = interp.mksimple_expr($2),
			 (delete $2, delete $3));
  $$ = interp.mklsect($3, x); }
| '(' simple LT ')'
{ expr *x; parser_action(x = interp.mksimple_expr($2),
			 (delete $2, delete $3));
  $$ = interp.mklsect($3, x); }
| '(' simple RT ')'
{ expr *x; parser_action(x = interp.mksimple_expr($2),
			 (delete $2, delete $3));
  $$ = interp.mklsect($3, x); }
| '(' simple PR ')'
{ expr *x; parser_action(x = interp.mksimple_expr($2),
			 (delete $2, delete $3));
  $$ = interp.mklsect($3, x); }

/* Right sections. */

| '(' NA simple ')'
{ expr *x; parser_action(x = interp.mksimple_expr($3),
			 (delete $2, delete $3));
  $$ = interp.mkrsect($2, x); }
| '(' LT simple ')'
{ expr *x; parser_action(x = interp.mksimple_expr($3),
			 (delete $2, delete $3));
  $$ = interp.mkrsect($2, x); }
| '(' RT simple ')'
{ expr *x; parser_action(x = interp.mksimple_expr($3),
			 (delete $2, delete $3));
  $$ = interp.mkrsect($2, x); }

;

comp_clauses
: comp_clause_list
| comp_clause_list ';'	{ $$ = $1; }
;

comp_clause_list
: comp_clause
{ $$ = new comp_clause_list; $$->push_back(*$1); delete $1; }
| comp_clause_list ';' comp_clause
{ $$ = $1; $$->push_back(*$3); delete $3; }
;

comp_clause
: expr
{ $$ = new comp_clause(*$1, expr()); delete $1; }
| expr '=' expr
{ $$ = new comp_clause(*$1, *$3); delete $1; delete $3; }
;

rows
: row_list
| row_list ';'	{ $$ = $1; }
| /* empty */	{ $$ = new exprll; }
;

row_list
: row
{ $$ = new exprll; $$->push_back(*$1); delete $1; }
| row_list ';' row
{ $$ = $1; $$->push_back(*$3); delete $3; }
;

row
: expr		{ $$ = interp.mkrow_exprl($1); }
;

/* Rewriting rule syntax. These generally take the form l = r [if g]; ... For
   convenience, we also allow a semicolon at the end of a rule list. Moreover,
   multiple left-hand sides are permitted (denoting a collection of rules for
   the same right-hand side), and the left-hand side may also be omitted, in
   which case the left-hand sides of the previous rule are repeated. */

rule
: lhs '=' rhs
{ $$ = new rulel;
  for (exprl::iterator l = $1->begin(), end = $1->end(); l != end; l++)
    $$->push_back(rule(*l, $3->rhs(), $3->qual()));
  delete $1; delete $3; }
| '=' rhs
{ $$ = new rulel(1, rule(expr(), $2->rhs(), $2->qual())); delete $2; }
;

lhs
: expr
{ $$ = new exprl; $$->push_back(*$1); delete $1; }
| lhs '|' expr
{ $$ = $1; $$->push_back(*$3); delete $3; }
;

rhs
: expr			{ $$ = new rhs_info($1); }
| qual_rhs
;

qual_rhs
: expr OTHERWISE	{ $$ = new rhs_info($1); }
| expr IF simple
{ expr *x; parser_action(x = interp.mksimple_expr($3),
			 (delete $1, delete $3));
  $$ = new rhs_info($1, x); }
| qual_rhs WHEN simple_rules END
{ $$ = $1;
  if ($$->q) {
    $$->r = interp.mkcond1_expr($$->q, $$->r); $$->q = 0;
  }
  try {
    expr *x = interp.mkwhen_expr($$->r, $3);
    $$->r = x;
  } catch (err &e) {
    interp.error(yyloc, e.what());
  }
}
| qual_rhs WITH rules END
{ $$ = $1;
  if ($$->q) {
    $$->r = interp.mkcond1_expr($$->q, $$->r); $$->q = 0;
  }
  expr *x = interp.mkwith_expr($$->r, new env($3->e)); delete $3;
  $$->r = x;
}
;

rules
: rulel
| rulel ';'		{ $$ = $1; }
;

rulel
: rule
{ $$ = new rule_info;
  rulel *rl = 0;
  try {
    rl = interp.default_lhs($$->l, $1);
    interp.add_rules($$->e, rl);
  }
  catch (err &e) { if (rl) delete rl; interp.error(yyloc, e.what()); } }
| rulel ';' rule
{ $$ = $1;
  rulel *rl = 0;
  try {
    rl = interp.default_lhs($$->l, $3);
    interp.add_rules($$->e, rl);
  }
  catch (err &e) { if (rl) delete rl; interp.error(yyloc, e.what()); } }
;

/* Same for pattern rules (pattern binding in 'case' clauses). */

pat_rules
: pat_rulel
| pat_rulel ';'		{ $$ = $1; }
;

pat_rulel
: rule
{ $$ = new pat_rule_info;
  rulel *rl = 0;
  try {
    rl = interp.default_lhs($$->l, $1);
    interp.add_rules($$->rl, rl, true);
  }
  catch (err &e) { if (rl) delete rl; interp.error(yyloc, e.what()); } }
| pat_rulel ';' rule
{ $$ = $1;
  rulel *rl = 0;
  try {
    rl = interp.default_lhs($$->l, $3);
    interp.add_rules($$->rl, rl, true);
  }
  catch (err &e) { if (rl) delete rl; interp.error(yyloc, e.what()); } }
;

/* Same for simple rules (pattern binding in 'when' clauses or 'let', 'const',
   'def', no guards in these cases). */

simple_rule
: expr '=' expr
{ $$ = new rule(*$1, *$3); delete $1; delete $3; }
| expr
{ expr *x = interp.mksym_expr(new string("_"));
  $$ = new rule(*x, *$1); delete x; delete $1; }
;

simple_rules
: simple_rulel
| simple_rulel ';'		{ $$ = $1; }
;

simple_rulel
: simple_rule
{ $$ = new rulel; try { interp.add_simple_rule(*$$, $1); }
  catch (err &e) { interp.error(yyloc, e.what()); } }
| simple_rulel ';' simple_rule
{ $$ = $1; try { interp.add_simple_rule(*$$, $3); }
  catch (err &e) { interp.error(yyloc, e.what()); } }
;

%%

void
yy::parser::error (const yy::parser::location_type& l,
		   const string& m)
{
  interp.error(l, m);
}

static void mangle_fname(string& name)
{
  size_t pos = name.find("::");
  while (pos != string::npos) {
    name.replace(pos, 2, "/");
    pos = name.find("::", pos);
  }
  name += ".pure";
}
